#!/usr/bin/env python
# coding: utf-8
#from qt import *
import threading
import re
import sys
from logging import debug, info, exception, error, warning
import time
import operator
import subprocess

class DcopAmaroK:
    """ convenience class to send dcop command to amarok"""
    class DCOPError(Exception):
        pass

    def __init__(self):
        self.message = ""

    def send(self, section, function, *args):
        cli= "amarok.%s.%s(%s)"%(
                section,
                function,
                ', '.join(map(repr, args))
        )

        debug("executing `%s'"%(cli))

        p = subprocess.Popen(
            ['dcop', 'amarok', section.strip(), function.strip()] + \
            [unicode(arg).strip() for arg in args],
            stdin = subprocess.PIPE , stdout=subprocess.PIPE
        )

        output = p.communicate()[0]
        if output:
            debug("got `%s' from amarok"%output.strip())
        if p.wait() != 0:
            error("Return code of %s was %s"%(cli, p.returncode))
            raise self.DCOPError
        return output.strip()
# file protocol to send a message to amarok "popupmessage" system
    def write(self, message):
        message = str(message)
        self.message += str("<p>"+message.replace("\n", "<br>")+"<\p>")

    def flush(self):
        if self.message:
            subprocess.Popen(
                ["dcop", "amarok", "playlist", "popupMessage", self.message],
                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        self.message = ''

    def close(self):
        self.write("exiting")
        self.flush()

class Amarok:
    def __init__(self):
        self.initLog()
        self._line=None
        self._finished = threading.Event()
        self.threads = []
        self.handlers={}
        self.addHandler('quit', self.quit)
        self.newThread(threading.Thread(
            target=self._readStdin, name="sdinReader Thread")
        )

    def initLog(self):
        """ set up default logging system.
        May be subclassed but self.log (a logging.getLoggerClass() or subclass instance)
        should be created """
        self.log = logging.getLogger('Amarok')
        formatter = logging.Formatter(
            "%(levelname)s in script %(module)s , thread %(threadName)s , (line %(lineno)d): %(message)s"
        )
        hdlr=logging.StreamHandler()
        hdlr.setFormatter(formatter)
        self.log.addHandler(hdlr)
        self.log.setLevel(logging.ERROR)
        self.log.debug('log initiated')

    def _readStdin(self):
        """Thread that look for events in the notifications from amarok """
        time.sleep(1)
        try:
            while not self.isFinished():
                line = sys.stdin.readline() #line includes a trailing \n
                if line:
                    l=line.strip()
                    if l:
                        self.customEvent(l)
                else:
                    self.log.info("got EOF")
                    break
        finally:
            self.quit()

    def customEvent(self, line):
        """This function search for an appropriate handler for the notification
        line and apply it (with args given to addHandler AND the keyword args
        set as the named group dict if supported)"""
        for event , (func, args) in self.handlers.items():
            m = re.match(event, line)
            if m:
                self.log.debug('%s match %s (with %s) '%(event.pattern, line, m.groupdict()))
                try:
                    func(*args, **m.groupdict())
                except TypeError:
                    self.log.exception("Warning, %s doesn't seem to provide kwargs suport"%func)
                    func(*args)

    def addHandler(self, event, func, args=[]):
        """event is a regexp that match an event from amarok.
        When a corresponding event is received (from stdin), the func is called with args as arguments and the kewords args corresponding to the named groups in regexp
        There is already a handler for quit: self.quit"""

        if not callable(func):
            raise TypeError, "%s is not callable"%(func, )
        self.handlers[re.compile(event)]=(func, args)

    def removeHandler(self, event):
        del self.handlers[re.compile(event)]

    def removeAllHandlers(self):
        self.handlers={"quit":(self.quit, [])}

    def newThread(self, thread):
        #XXX: is MainThread always the first in enumerate?
        ct = threading.currentThread()
        if ct is not threading.enumerate()[0]:
            self.log.warning("nested thread %s in %s"%(thread, ct))
        thread.start()
        self.threads.append(thread)
        self._compactThreads()
        return thread

    def _compactThreads(self):
        self.threads = filter(lambda t:t.isAlive(), self.threads)
        self.log.debug(self.threads)

    def wait(self, *args):
        self.log.debug("waiting for the finished flag")
        self._finished.wait(*args)
        if not self._finished.isSet():
            self.log.debug("timeout ended whitout exiting")
            return False
        for t in self.threads:
            self.log.debug(threading.enumerate())
            if  t is not threading.currentThread() and t.isAlive():
                self.log.debug("Joining:%s"%t)
                t.join(*args)
                self.log.debug("done")
        return True

    def isFinished(self):
        """return True if quit() has been called at least once
        Warning: it doesn't mean that all threads has finished
        """
        return self._finished.isSet()

    def quit(self):
        """Stop all internal threads.
        This should be subclassed and called explicitly (Amarok.quit(self))
        if these are things to clean in the subclass instance"""
        self.log.debug("setting finished flag")
        self._finished.set()
        self.wait()
