#!/usr/bin/env python
#coding:utf-8
############################################################################
# Config dialog for weekalarm script
# (c) 2008 Maxime Steisel <maximesteisel@gmail.com>
# (c) 2007 Brady Laska <bfl@clarkhall.ca>
#
# Depends on: Python 2.2, PyQt
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################


#from __future__ import with_statement

import locale
try:
    locale.setlocale(locale.LC_ALL, '')
except:
    pass

import threading
from logging import debug, info, exception, error, warning, handlers
import logging
import datetime
import random
import subprocess
import time
import sys
import ConfigParser

import gettext
gettext.install('weekalarm', "%s/l10n"%sys.path[0])

import amarok
from config import * # ConfigDialog and globals like days + qt,...

class WeekAlarm( amarok.Amarok, qt.QApplication ):
    class Config(object):
        def __init__(self, config):
            self.__dict__.update({"config":config})

        def __getattr__(self, attr):
#called only if attr not in __dict__
            r = self.config[attr]
            if r == {}:
                error('self.%s not found'%attr)
                raise KeyError
            return r
        def __setattr__(self, attr, val):
            if attr not in self.config.shorts:
                return self.__dict__.update({attr:val})
            self.config[attr] = val
        def write(self):
            self.config.write()

    def __init__(self):
        qt.QApplication.__init__(self, sys.argv)

        self.dcop = amarok.DcopAmaroK()

        amarok.Amarok.__init__(self)

        self.pause = threading.Event()
        self.fadeLock = threading.Lock()
        self.wakeLock = threading.Lock()
        self.snoozeLock = threading.Lock()
#we need a self.timer attr and it must be started to eventually be joined
        self.timer = self.newThread(
                threading.Timer(0, lambda:None)
                )
        self.currentVol = int(
                self.dcop.send("player", "getVolume")
                )
        self.dialog = ConfigDialog()
        self.configuration = qt.PYSIGNAL("configuration()")
        self.connect(self, self.configuration, self.dialog.run)
        self.config = self.Config(self.dialog.config)
        self.connect(self.dialog, qt.PYSIGNAL("accepted()"), self.readSettings)
        self.readSettings()

        self.initHandlers()

    def initHandlers(self):
        self.addHandler(r"^configure$", self.emit, (self.configuration, ()))
        if '-v' in sys.argv[1:]:
#testing handler
            self.addHandler(r"^eval\((?P<line>.*)\)$", self.eval)
        self.addHandler(
            r"^engineStateChange\s*:\s*(paused)$",
            lambda: self.config.stopOnPause and self.pause.set()
        )
        self.addHandler(
            r"^engineStateChange\s*:\s*(empty)$",
            lambda: self.config.stopOnPause and self.pause.set()
        )
        self.addHandler(
            r"volumeChange: (?P<newvol>\d*)",
            self.volumeHandler
            )
        self.addHandler(r"^engineStateChange\s*:\s*playing$", self.pause.clear)

    def initLog(self):
        self.log = logging.getLogger('Amarok.WeekAlarm')
        hdlr = logging.StreamHandler(self.dcop)
        hdlr.setLevel(logging.ERROR)
        fmt = logging.Formatter("Script WeekAlarm: %(message)s")
        hdlr.setFormatter(fmt)
        self.log.addHandler(hdlr)

    def eval(self, line):
        self.log.debug("evaluating line:%s"%line)
        try:
            self.log.debug("eval returned %s"%eval(line))
        except Exception:
            self.log.exception("got an exception evaluating %s"%line)

    def quit(self):
        if not self.isFinished():
            self.log.debug("exiting")
        self.timer.cancel()
        self.exit_loop()
        amarok.Amarok.quit(self)

    def wait(self, *args):
        if amarok.Amarok.wait(self, *args):
            self.log.debug("amarok wait returned, joining timer")
            self.timer.join()

    def readSettings(self):
        self.log.debug("weekalarm.read")

        self.config.daysOn = []
        self.config.alarmTimes = []
        for day in days:
            self.config.daysOn.append(getattr(self.config, day))
            self.config.alarmTimes.append(getattr(self.config, day + 'Time'))

        self.setTimer()

    def setTimer(self, time = None):
        try:
            if time is None:
                time = self.nextAlarm()
            self.timer.cancel() #cancelling previous timer
            self.log.debug("Setting timer to %s seconds from now"%time)
            if time is not None:
                if not (isinstance(time, int) or isinstance(time, float)):
                    raise TypeError, ('Arg time must be an int or None, '
                        'not %s')%(type(time))
                if time < 0:
                    raise ValueError, "Argument time can't be negative"
                self.timer = threading.Timer(time, self.wakeup)
                self.newThread(self.timer)
                return self.timer
        except (TypeError, ValueError), e:
            raise
        except Exception, e:
            self.log.exception("Got an error while setting timer:")
            self.quit()
            raise

    def nextAlarm(self):
        if self.config.napMode:
            self.log.info("Next alarm in %s minute"%(self.config.napLen, ))
            self.config.napMode = False
            self.config.write()
            return self.config.napLen * 60
        now = datetime.datetime.now()
        for i in range(0, 8): #it can be today if the conf file changed
            date = now.date() + datetime.timedelta(days = i)

            time_qt = self.config.alarmTimes[date.weekday()]
# QTime.toPyTime only available since pyqt4
# .toString return q QString
            time_tuple = map(int, str(time_qt.toString()).split(':'))
            then_time = datetime.time(*time_tuple)

            then = datetime.datetime.combine(
                date, then_time
                )
# depending on the time, then could be < now
            if then > now and self.config.daysOn[then.weekday()]:
                self.log.info("Next alarm: %s"%then)
                mkt = time.mktime
                return mkt(then.timetuple())-mkt(now.timetuple())

    def wakeup( self ):
        try:
            try:
                if not self.wakeLock.acquire(False):
                    self.log.info("Wakeup cannot acquire the lock. Is it"
                    " already playing ?")
                    return
                self.pause.clear()
                self.log.info("wakeup")
                # disable nap mode and resume regular alarms
                self.config.napMode = False

                # load the playlist if required
#WARNING: this doesn't always work
                if self.config.plMode != PL_NO:
                    plIndex = '0'
                    self.dcop.send('playlist', 'clearPlaylist')
                    if self.config.plMode == PL_DYNAMIC:
# load playlist is not very reliable, so trying some time before give up
                        t = 0
                        while t < 5:
                            exitcode = self.dcop.send('playlistbrowser', 'loadPlaylist',
                                       self.config.pl)
                            if int(exitcode):
                                t+=1
                            else:
                                self.log.debug("sucess after %s tries"%t)
                                break
                    else:
                        self.dcop.send('playlist', 'addMedia', self.config.pl)
                        exitcode = 0
                    if int(exitcode):
                        self.log.debug(exitcode)
                        raise Exception("Cannot load %s, please check the"
                                " name"%self.config.pl)
                    # make sure the playlist gets fully populated,
                    # othewise the player won't start
                    i = 0
                    while int(
                        self.dcop.send("playlist", "getTotalTrackCount")
                    ) == 0 and i < 15:
                        time.sleep(1)
                        i +=1 #avoid infinite recursion if playlist is empty
                elif self.config.volMode != VOL_DECREASE:
# if a custom playlist isn't being used,
# and mode is not "decrease volume and stop",
# get the current track (default to 0)

                    plIndex = int(
                        self.dcop.send("playlist", "getActiveIndex")
                    )
                    if plIndex == -1: plIndex = 0
                    self.log.debug("Current track:%s"%plIndex)
                self.dcop.send("player", "enableRepeatTrack",
                               self.config.repeat)
                self.dcop.send("player", "enableRandomMode",
                               self.config.random)
                if self.config.random is True:
                    plIndex = random.randrange(
                        0, int(self.dcop.send("playlist", "getTotalTrackCount"))
                    )
                # handle the volume fading and snooze
                if self.config.volMode == VOL_NORMAL:
                    self.dcop.send('playlist', 'playByIndex', plIndex)
                    self.newThread(
                        threading.Thread(target=self.checkSnooze)
                    )
                elif self.config.volMode  == VOL_DECREASE:
                    v = int(self.dcop.send("player", "getVolume"))
                    self.fadeVol(v, -1, -1)
                    self.dcop.send('player', 'stop')
                elif self.config.volMode == VOL_INCREASE:
                    self.dcop.send(
                        'player', 'setVolume',
                        str(self.config.fadeStart)
                    )
                    self.dcop.send('playlist', 'playByIndex', plIndex)
                    self.newThread(threading.Thread(target=self.checkSnooze))
                    self.fadeVol(self.config.fadeStart,
                                 self.config.fadeEnd+1, 1)
            except Exception:
                self.log.exception(
                    """There were a problem when launching alarm.
    Please report this bug (with traceback) to the author (see readme)"""
                )
                raise
        finally:
            self.wakeLock.release()
            self.log.debug("reset Timer")
            self.timer.cancel()
            if not (self.isFinished()):
                self.setTimer()

    def checkSnooze(self):
        if self.config.snooze:
            # kdialog returns False if the answer was yes
            if self.snoozeLock.acquire(False): #acquire without block
                dontSnooze= bool(
                    subprocess.call('kdialog --yesno "%s?"'%_("snooze"), shell=True)
                )
                self.snoozeLock.release()
            else:
                return

            self.log.debug("dontSnooze : %s"%dontSnooze)
            if not(dontSnooze):
                self.dcop.send('player', 'pause')
                self.pause.set()
                self.log.debug(
                    "sleeping %s seconds"%(60*self.config.snoozeLen, )
                )
                f = self.wait(60*self.config.snoozeLen)
                self.log.debug("Snooze ended")
                if not f:
                    self.wakeup()

    def volumeHandler(self, newvol):
        newvol = int(newvol)
        if newvol != self.currentVol:
            self.log.debug("Volume manually changed")
            self.log.debug("%s -> %s (t:%s/%s)"%(self.currentVol,
                newvol,
                self.config.VolumeSens,
                self.config.stopVolume,
                )
            )
            if self.config.stopVolume:
                if abs(newvol - self.currentVol) > self.config.VolumeSens:
                    if not self.pause.isSet():
                        self.log.info("Volume has changed,"
                                      "this will stop fading")
                        self.pause.set()
            self.currentVol = newvol

    def fadeVol(self, start, stop, step):
        if self.fadeLock.acquire(False):
            i = start
            while i < min(stop, 101):
                if self.isFinished() or self.pause.isSet():
                    self.log.debug("Canceling fade")
                    break
                self.currentVol = i
                self.dcop.send('player', 'setVolume', str(i) )
                time.sleep(self.config.fadeTime)
                if self.config.updateVol:
                    if self.config.updateEnd:
                        stop += self.currentVol - i
                        self.log.debug("stop = %s"%stop)
                    i = self.currentVol
                i += step
            self.fadeLock.release()

    def saveState(self, sessionmanager):
        # script is started by amarok, not by KDE's session manager
        sessionmanager.setRestartHint(qt.QSessionManager.RestartNever)

def initLog():

    root = logging.getLogger('')
    root.setLevel(logging.DEBUG)
    console = logging.StreamHandler(sys.stderr)
    if '-v' in sys.argv:
        level = logging.DEBUG
    else:
        level = logging.INFO
    console.setLevel(level)
    console.setFormatter(logging.Formatter(
        "%(asctime)s %(message)s [%(levelname)s]",
        datefmt="%X"
        )
    )
    root.addHandler(console)

    #log:changes every days, 3 files max
    flog=handlers.TimedRotatingFileHandler('weekalarm.log', "d", 1, 3)
    flog.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s %(message)s (in %(threadName)s) [%(levelname)s]",
        datefmt="%x %X",
    )
    flog.setFormatter(formatter)
    root.addHandler(flog)

def main():
    initLog()
    info("weekalarm beginning")
    app=WeekAlarm()
    app.exec_loop()
    info("weekalarm finished")
    debug(threading.enumerate())


if __name__ == '__main__':
    main()
